
/* Based on code inspiration from
	http://www.freesoftwaremagazine.com/articles/drivers_linux?page=0%2C11 
	http://morethanuser.blogspot.no/2013/04/raspbery-pi-gpio-interrupts-in-kernel.html
*/ 


#include <linux/init.h>

#include <linux/module.h>
#include <linux/kernel.h> /* printk() */

#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/gpio.h>

#include <linux/slab.h>   /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>  /* O_ACCMODE */
// #include <asm/system.h>   /* cli(), *_flags */
#include <asm/uaccess.h>  /* copy_from/to_user */

MODULE_LICENSE("Dual BSD/GPL");

#define A_GBLEN 100

#define GPIO_PIN_STR "Accumulator GPIO input"
#define GPIO_PIN_DEV_STR "GPIO"
#define THISFILE "accumulator.c"

/* Declaration of accumulator.c functions */

static int a_open(struct inode *inode, struct file *filp);
static int a_release(struct inode *inode, struct file *filp);
static ssize_t a_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
static ssize_t a_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos);
static void a_exit(void);
static int a_init(void);

/* Structure that declares the access functions. */

struct file_operations accumulator_fops = 
{
  read: a_read,
  write: a_write,
  open: a_open,
  release: a_release
};

/* Declaration of the init and exit functions */
module_init(a_init);
module_exit(a_exit);

/* enforce opening only once. */
static int a_open_count = 0;

/* Major number */
static int a_major = 0;


/* Global variables of the driver. */

/* Which irq is being used (automatically defined) */ 
short int accumulator_irq = 0;
short int accumulator_irq_enabled = 0; 

/* State variables */
unsigned short accumulator_ticks = 0;

/* Accumulator rate values: every time, we add A, then for every 
	N ticks we add an additional D. Will allow all unit divisors 
	up to 65536 and many values above this to be used. */ 

unsigned short accumulator_A;
unsigned short accumulator_N;
unsigned short accumulator_D;

unsigned long long accumulator_value = 0;

/* Module params */

/* Which gpio line to be used */
static short gpio_line = 22;

/* Accumulator A, N, D settings. */

/* These default values match the 1p = 2/3 m of the wind speed sensor */
static int acc_A = 0xaaaa;
static int acc_N = 3;
static int acc_D = 2;
/* Use the ratecalc program to figure out the correct values. */

module_param(gpio_line, short, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP); 
MODULE_PARM_DESC(gpio_line, GPIO_PIN_STR);

module_param(acc_A, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP); 
MODULE_PARM_DESC(acc_A, "Accumulator factor A");

module_param(acc_N, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP); 
MODULE_PARM_DESC(acc_N, "Accumulator additionals rate N");

module_param(acc_D, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP); 
MODULE_PARM_DESC(acc_D, "Accumulator additional D");

/* IRQ handler */
static irqreturn_t a_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long flags;

	local_irq_save(flags); 

	/* If the irq isn't enabled we don't do anything ! */ 
	if(accumulator_irq_enabled)
	{
		// printk(KERN_INFO "%s(%d): Irq is active\n", THISFILE, __LINE__); 
		accumulator_value += accumulator_A;
		accumulator_ticks ++;
		if(accumulator_ticks >= accumulator_N )
		{
			accumulator_value += accumulator_D; 
			accumulator_ticks = 0;
		}
	}
	else
	{
		printk(KERN_INFO "%s(%d): Irq not active\n", THISFILE, __LINE__); 
	}
	local_irq_restore(flags); 
	return IRQ_HANDLED;
}

/* Init: enable interrupt, create the device entry. */
static int a_init(void) 
{
	int rx;

	accumulator_A = acc_A;
	accumulator_N = acc_N;
	accumulator_D = acc_D;

	/* Request the gpio line, enable the interrupt on it. */
	rx = gpio_request( gpio_line, GPIO_PIN_STR);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d request failure %d\n", THISFILE, __LINE__, gpio_line, rx); 
		return(rx);
	}

	accumulator_irq = gpio_to_irq(gpio_line);
	if(accumulator_irq < 0)
	{
		printk(KERN_INFO 
			"%s(%d): GPIO %d to IRQ mapping failure %d\n", THISFILE, __LINE__, 
			gpio_line, accumulator_irq); 

		gpio_free(gpio_line); 
		return(accumulator_irq);
	}

	printk(KERN_INFO "%s(%d): Mapped int %d from gpio-line %d\n", THISFILE, __LINE__, 
			accumulator_irq, gpio_line); 

	rx = request_irq(accumulator_irq, 
			(irq_handler_t) a_irq_handler, 
			IRQF_TRIGGER_FALLING, /* IRQF_TRIGGER_RISING is an alternative */
			GPIO_PIN_STR,
			GPIO_PIN_DEV_STR); 

	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): Irq %d trigger-request on gpio %d failure %d\n", THISFILE, __LINE__, 
			accumulator_irq, 
			gpio_line, rx); 
		gpio_free(gpio_line); 
		return(rx);
	}

	printk(KERN_INFO "%s(%d): Info: gpio-line %d\n", THISFILE, __LINE__, gpio_line); 
	
	/* Registering the device itself. */
	a_major = register_chrdev(0, "accumulator", &accumulator_fops);

	if(a_major < 0) 
	{
		printk(KERN_INFO 
			"%s(%d): accumulator: cannot obtain major number, failure= %d\n", THISFILE, __LINE__, 
			a_major);

		free_irq(accumulator_irq, GPIO_PIN_DEV_STR); 
		gpio_free(gpio_line); 
		return a_major;
	}

	printk(KERN_INFO "%s(%d): Inserting accumulator module major no %d\n", THISFILE, __LINE__, a_major); 
	printk(KERN_INFO "%s(%d): Now do \'mknod /dev/accumulator c %d 0\'\n", THISFILE, __LINE__, a_major); 
	return 0;
}


static void a_exit(void) 
{
	free_irq(accumulator_irq, GPIO_PIN_DEV_STR); 
	gpio_free(gpio_line); 
	unregister_chrdev(a_major, "accumulator");
	printk(KERN_INFO "%s(%d): Removing accumulator module\n", THISFILE, __LINE__);
}


static int a_open(struct inode *inode, struct file *filp) 
{
	if(a_open_count)
		return -EBUSY;

	a_open_count ++;
	try_module_get(THIS_MODULE); 

	return 0;
}

static int a_release(struct inode *inode, struct file *filp) 
{
	a_open_count --;
	module_put(THIS_MODULE);
	return 0;
}


static ssize_t a_read(struct file *filp, char *buff_out, size_t count, loff_t *f_pos) 
{ 
	size_t wcount;
	char wbuff[A_GBLEN];

	/* Accumulator-value is really a number that we should
		divide by 65536 in order to get the main 
		unit (km, L, mm-rain, whatever) */

	wcount = snprintf(wbuff, sizeof(wbuff), "%llu\n", accumulator_value);

	/* Transfering data to user space */ 
	copy_to_user(buff_out, wbuff, wcount);

	/* Changing reading position as best suits */ 
	if(*f_pos == 0) 
	{ 
		*f_pos += wcount; 
		return wcount; 
	} 
	else 
	{ 
		return 0; 
	}
}



/* Write -- we come in with some action from the outside,
	write new A, D, N values, start/stop the isr, or 
	reset the acc.  (should probably use ioctls instead for these? but
	that requires a fixed major number.)

	commands:
		start - enables the irq
		stop - disables the irq
		reset - sets accumulator to 0
*/

static ssize_t a_write(struct file *filp, const char *buff_in, size_t count, loff_t *f_pos) 
{
	char rdbuff[A_GBLEN];

	if(count > A_GBLEN)
	{
		printk(KERN_INFO "%s(%d): Read error buffer overflow %d < %d\n", THISFILE, __LINE__, count, A_GBLEN); 
		return(-EINVAL);
	}
	
	copy_from_user(rdbuff, buff_in, count);
	rdbuff[count] = 0;
	if(rdbuff[count-1] == '\n') rdbuff[count-1] = 0; 

	rcu_read_unlock(); /* Must use or code will loop ? */ 

	if(!strcmp("start", rdbuff))
	{
		accumulator_irq_enabled = 1; 
	}
	else
	if(!strcmp("stop", rdbuff))
	{
		accumulator_irq_enabled = 0; 
	}
	else
	if(!strcmp("reset", rdbuff))
	{
		accumulator_value = 0;
	}
	else
	{
		printk(KERN_INFO "%s(%d): Written unknown command %s\n", THISFILE, __LINE__, rdbuff); 
	}
	// set A, D, N
	// zero-out
	// Look at the stuff that the user wrote... 
	// enable/disable the irq
	// accumulator_irq_enabled = 1; 


	return(count); 
}



