
/* Based on code inspiration from
	http://www.freesoftwaremagazine.com/articles/drivers_linux?page=0%2C11 
	http://morethanuser.blogspot.no/2013/04/raspbery-pi-gpio-interrupts-in-kernel.html
	http://www.linuxvoice.com/be-a-kernel-hacker/
*/ 

/* Rpi gamma pushbutton:
	gpio 18 (P1.12) - H when pushed, L when not pushed
	gpio 22 (P1.15) - I channel
	gpio 23 (P1.16) - Q channel
*/ 

/* Define using three of the real-time signals:
	SIGRTMIN+0 - Pushed
	SIGRTMIN+1 - Deasil movement
	SIGRTMIN+2 - Widdershins movement

	Check that we're never exceeding SIGRTMAX.
*/

#include <linux/init.h>
#include <linux/miscdevice.h>

#include <linux/module.h>
#include <linux/kernel.h> /* printk() */

#include <linux/interrupt.h>
#include <linux/gpio.h>

#include <linux/signal.h>  /* Signal numbers etc. */
#include <linux/sched.h>  /* process-ident things. */

#include <linux/slab.h>   /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>  /* O_ACCMODE */
// #include <asm/system.h>   /* cli(), *_flags */
#include <asm/uaccess.h>  /* copy_from/to_user */

#include "gammapb_u.h"  /* Signal numbers. */

/* GPIO space */
#include <mach/platform.h>

MODULE_LICENSE("Dual BSD/GPL");

#define A_GBLEN 100

#define PB_LABEL "Pushbutton"
#define IC_LABEL "I Channel"
#define QC_LABEL "Q Channel"

#define GPIO_PIN_DEV_STR "GAMMAPB"
#define THISFILE "gammapb.c"

/* Declaration of gammapb.c functions */

static int g_open(struct inode *inode, struct file *filp);
static int g_release(struct inode *inode, struct file *filp);
static ssize_t g_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
static ssize_t g_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos);
static void g_exit(void);
static int g_init(void);

/* Structure that declares the access functions. */

struct file_operations gammapb_fops = 
{
  read: g_read,
  write: g_write,
  open: g_open,
  release: g_release
};

/* Misc device structure. */
static struct miscdevice gammapb_misc_device 
	= {
	.minor = MISC_DYNAMIC_MINOR,
	.name = "gammapb",
	.fops = &gammapb_fops 
  };

/* Declaration of the init and exit functions */
module_init(g_init);
module_exit(g_exit);

/* enforce opening only once. */
static int g_open_count = 0;

/* Which signal to send and where to send it? */
pid_t send_to_pid = 0; /* Set in write or better, ioctl. */

short int pb_signal = SIG_PB; /* SIGRTMIN; */ 
module_param(pb_signal, short, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP ); 
MODULE_PARM_DESC(pb_signal, "Signal to send on push"); 

short int deasil_signal = SIG_DEASIL; /* SIGRTMIN+1; */ 
module_param(deasil_signal, short, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP ); 
MODULE_PARM_DESC(deasil_signal, "Signal to send on deasil movement"); 

short int wshins_signal = SIG_WIDDERSHINS; /* SIGRTMIN+2; */ 
module_param(wshins_signal, short, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP ); 
MODULE_PARM_DESC(wshins_signal, "Signal to send on widdershins movement"); 

/* Global variables of the driver. */

/* Which irqs are being used (automatically defined) */ 
static short int gammapb_pb_irq  = 0;
static short int gammapb_ic_irq  = 0;
static short int gammapb_qc_irq  = 0;

static short int gammapb_irq_enabled = 0; 

static unsigned int gammapb_count = 0;
static int gammapb_knob = 0;

/* GPIOs being handled here */
static short gpio_pb = 18;
static short gpio_ic = 22;
static short gpio_qc = 23;

/* Send the indicated signal if the send_to_pid is nonzero and oke */ 
int sendasignal(int id, short sigtosend)
{
	int ssig;
	struct task_struct *task_from_pid;

	printk(KERN_INFO "%s(%d): %c: About to send signal %d \n", THISFILE, __LINE__, id, sigtosend); 

	/* The following sends the signal to the 
		listening process, if there is one. */ 
	if(send_to_pid)
	{
		task_from_pid = pid_task(find_vpid(send_to_pid), PIDTYPE_PID); 

		if(task_from_pid == NULL) 
		{
			// no-one to send a signal to
			return(0);
		}

		ssig = send_sig_info(sigtosend, SEND_SIG_FORCED, task_from_pid);
		if(ssig < 0)
		{
			// signal wasn't sent
			return(0);
		}

		// signal has been sent
		return(ssig);
	}
	// signal not sent as there were noone to send it to 
	return(0);
}


/* IRQ handlers */

/* I channel irq is detected: I-channel has an edge, the state of the Q channel
	will indicate direction forward or reverse. */
static irqreturn_t g_irq_ic_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long flags;
	int qv;

	local_irq_save(flags); 

	if(!gammapb_irq_enabled)
	{
		printk(KERN_INFO "%s(%d): Irq not active\n", THISFILE, __LINE__); 
		goto theirqisdone;
	}

	/* printk(KERN_INFO "%s(%d): Irq is active\n", THISFILE, __LINE__);  */ 

	/* Look at the Q channel gpio line, */
	qv = gpio_get_value(gpio_qc);
	/* printk(KERN_INFO "%s(%d): IC: Q-channel = %d\n", THISFILE, __LINE__, qv);  */ 

	if(qv == 0) /* Deasil */
	{
		gammapb_knob++;
		sendasignal('d', deasil_signal); 
	}
#if(0)
	else /* Widdershins */
	{
		gammapb_knob--;
		sendasignal('w', wshins_signal); 
	}
#endif

theirqisdone:

	local_irq_restore(flags); 
	return IRQ_HANDLED;
}

static irqreturn_t g_irq_qc_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long flags;
	int iv;
	local_irq_save(flags); 

	if(!gammapb_irq_enabled)
	{
		printk(KERN_INFO "%s(%d): Irq not active\n", THISFILE, __LINE__); 
		goto theirqisdone;
	}

	/* printk(KERN_INFO "%s(%d): Irq is active\n", THISFILE, __LINE__);  */ 

	/* Look at the I channel gpio line, */
	iv = gpio_get_value(gpio_ic);
	/* printk(KERN_INFO "%s(%d): QC: I-channel = %d\n", THISFILE, __LINE__, iv);  */ 

	if(iv == 0) /* Widdershins */
	{
		gammapb_knob--;
		sendasignal('w', wshins_signal); 
	}
#if(0)
	else /* Deasil */
	{
		gammapb_knob++;
		sendasignal('d', deasil_signal); 
	}
#endif

theirqisdone:

	local_irq_restore(flags); 
	return IRQ_HANDLED;
}

static irqreturn_t g_irq_pb_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long flags;
	local_irq_save(flags); 

	/* If the irq isn't enabled we don't do anything ! */ 
	if(!gammapb_irq_enabled)
	{
		printk(KERN_INFO "%s(%d): Irq not active\n", THISFILE, __LINE__); 
		goto theirqisdone;
	}

	// printk(KERN_INFO "%s(%d): Irq is active\n", THISFILE, __LINE__); 
	gammapb_count += 1;

	sendasignal('p', pb_signal); 

theirqisdone:

	local_irq_restore(flags); 
	return IRQ_HANDLED;
}


/* use:
	gpio_direction_input(gpio_pb);
	value = gpio_get_value(gpio_pb);

	or: 
	gpio_direction_output(gpio_pb, value);
	gpio_set_value(gpio_pb, value);
*/ 


/* Init: enable interrupt, create the device entry. */
static int g_init(void) 
{
	int rx;

	/* Request the gpio lines and enable the interrupts on them. */
	rx = gpio_request( gpio_pb, PB_LABEL);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d request failure %d\n", 
			THISFILE, __LINE__, gpio_pb, rx); 
		goto nopbgpio;
	}

	rx = gpio_direction_input(gpio_pb);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d input direction request failure %d\n", 
			THISFILE, __LINE__, gpio_pb, rx); 
		goto nopbinput;
	}

	rx = gpio_request( gpio_ic, IC_LABEL);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d request failure %d\n", 
			THISFILE, __LINE__, gpio_ic, rx); 
		goto noicgpio;
	}

	rx = gpio_direction_input(gpio_ic);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d input direction request failure %d\n", 
			THISFILE, __LINE__, gpio_ic, rx); 
		goto noicinput;
	}

	rx = gpio_request( gpio_qc, QC_LABEL);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d request failure %d\n", 
			THISFILE, __LINE__, gpio_qc, rx); 
		goto noqcgpio;
	}

	rx = gpio_direction_input(gpio_qc);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d input direction request failure %d\n", 
			THISFILE, __LINE__, gpio_qc, rx); 
		goto noqcinput;
	}

	/* Set the irqs */
	gammapb_pb_irq = gpio_to_irq(gpio_pb);
	if(gammapb_pb_irq < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d to IRQ mapping failure %d\n", 
			THISFILE, __LINE__, gpio_pb, gammapb_pb_irq); 
		rx = gammapb_pb_irq;
		goto nopbirq;
	}

	printk(KERN_INFO "%s(%d): Mapped int %d from gpio-line %d\n", 
			THISFILE, __LINE__, gammapb_pb_irq, gpio_pb); 

	rx = request_irq(gammapb_pb_irq, 
			(irq_handler_t) g_irq_pb_handler, 
			IRQF_TRIGGER_FALLING, /* IRQF_TRIGGER_RISING is an alternative */
			PB_LABEL,
			GPIO_PIN_DEV_STR); 

	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): Irq %d trigger-request on gpio %d failure %d\n", THISFILE, __LINE__, gammapb_pb_irq, gpio_pb, rx); 
		goto nopbirq;
	}

	/* I channel irq */
	gammapb_ic_irq = gpio_to_irq(gpio_ic);
	if(gammapb_ic_irq < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d to IRQ mapping failure %d\n", 
			THISFILE, __LINE__, gpio_ic, gammapb_ic_irq); 
		rx = gammapb_ic_irq;
		goto noicirq;
	}

	printk(KERN_INFO "%s(%d): Mapped int %d from gpio-line %d\n", 
			THISFILE, __LINE__, gammapb_ic_irq, gpio_ic); 

	rx = request_irq(gammapb_ic_irq, 
			(irq_handler_t) g_irq_ic_handler, 
			IRQF_TRIGGER_FALLING, /* IRQF_TRIGGER_RISING is an alternative */
			IC_LABEL,
			GPIO_PIN_DEV_STR); 

	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): Irq %d trigger-request on gpio %d failure %d\n", 
			THISFILE, __LINE__, gammapb_ic_irq, gpio_ic, rx); 
		goto noicirq;
	}

	/* Q channel irq */
	gammapb_qc_irq = gpio_to_irq(gpio_qc);
	if(gammapb_qc_irq < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d to IRQ mapping failure %d\n", 
			THISFILE, __LINE__, gpio_qc, gammapb_qc_irq); 
		rx = gammapb_qc_irq;
		goto noqcirq;
	}

	printk(KERN_INFO "%s(%d): Mapped int %d from gpio-line %d\n", 
			THISFILE, __LINE__, gammapb_qc_irq, gpio_qc); 

	rx = request_irq(gammapb_qc_irq, 
			(irq_handler_t) g_irq_qc_handler, 
			IRQF_TRIGGER_FALLING, /* IRQF_TRIGGER_RISING is an alternative */
			QC_LABEL,
			GPIO_PIN_DEV_STR); 

	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): Irq %d trigger-request on gpio %d failure %d\n", 
			THISFILE, __LINE__, gammapb_qc_irq, gpio_qc, rx); 
		goto noqcirq;
	}

	printk(KERN_INFO "%s(%d): Info: gpio-line %d\n", THISFILE, __LINE__, gpio_pb); 
	printk(KERN_INFO "%s(%d): Info: gpio-line %d\n", THISFILE, __LINE__, gpio_ic); 
	printk(KERN_INFO "%s(%d): Info: gpio-line %d\n", THISFILE, __LINE__, gpio_qc); 

	/* Registering the device itself. */
	rx = misc_register(&gammapb_misc_device);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): Register misc device failure %d\n", THISFILE, __LINE__, rx); 
		goto nodevice;
	}

	printk(KERN_INFO "%s(%d): Info: misc device minor number = %d\n", 
		THISFILE, __LINE__, gammapb_misc_device.minor);

	return 0;

/* Error exits */
nodevice:
	free_irq(gammapb_qc_irq, GPIO_PIN_DEV_STR); 
noqcirq:
	free_irq(gammapb_ic_irq, GPIO_PIN_DEV_STR); 
noicirq:
	free_irq(gammapb_pb_irq, GPIO_PIN_DEV_STR); 
nopbirq:
noqcinput:
	gpio_free(gpio_qc); 
noqcgpio:
noicinput:
	gpio_free(gpio_ic); 
noicgpio:
nopbinput:
	gpio_free(gpio_pb); 
nopbgpio:
	return(rx);
}


static void g_exit(void) 
{
	free_irq(gammapb_qc_irq, GPIO_PIN_DEV_STR); 
	free_irq(gammapb_ic_irq, GPIO_PIN_DEV_STR); 
	free_irq(gammapb_pb_irq, GPIO_PIN_DEV_STR); 
	gpio_free(gpio_qc); 
	gpio_free(gpio_ic); 
	gpio_free(gpio_pb); 
	misc_deregister(&gammapb_misc_device);
	printk(KERN_INFO "%s(%d): Removing gammapb\n", THISFILE, __LINE__);
}


static int g_open(struct inode *inode, struct file *filp) 
{
	if(g_open_count)
		return -EBUSY;

	g_open_count ++;
	try_module_get(THIS_MODULE); 

	return 0;
}

static int g_release(struct inode *inode, struct file *filp) 
{
	g_open_count --;
	module_put(THIS_MODULE);
	return 0;
}


static ssize_t g_read(struct file *filp, char *buff_out, size_t count, loff_t *f_pos) 
{ 
	size_t wcount;
	char wbuff[A_GBLEN];

	/* Read: tbd, we will want to return a count for the knob turns. */

	wcount = snprintf(wbuff, sizeof(wbuff), 
		"%u %d\n", gammapb_count, gammapb_knob );

	/* Transfering data to user space */ 
	copy_to_user(buff_out, wbuff, wcount);

	/* Changing reading position as best suits */ 
	if(*f_pos == 0) 
	{ 
		*f_pos += wcount; 
		return wcount; 
	} 
	else 
	{ 
		return 0; 
	}
}



/* Write -- we come in with some action from the outside,

	commands:
		start - enables the irq
		stop - disables the irq
		reset - sets gammapb to 0
		Pnnnn - where nnnn is a process id, pid 
			- start sending signals to the indicated pid 
*/

static ssize_t g_write(struct file *filp, const char *buff_in, size_t count, loff_t *f_pos) 
{
	char rdbuff[A_GBLEN];
	int skf;

	if(count > A_GBLEN)
	{
		printk(KERN_INFO "%s(%d): Read error buffer overflow %d < %d\n", THISFILE, __LINE__, count, A_GBLEN); 
		return(-EINVAL);
	}
	
	copy_from_user(rdbuff, buff_in, count);
	rdbuff[count] = 0;
	if(rdbuff[count-1] == '\n') rdbuff[count-1] = 0; 

	rcu_read_unlock(); /* Must use or code will loop ? */ 

	if(!strcmp("start", rdbuff))
	{
		gammapb_irq_enabled = 1; 
	}
	else
	if(!strcmp("stop", rdbuff))
	{
		gammapb_irq_enabled = 0; 
	}
	else
	if(!strcmp("reset", rdbuff))
	{
		gammapb_knob = 0;
		gammapb_count = 0;
	}
	else /* Pnnnn for defining the process number to send signals to */
	if(rdbuff[0] == 'P')
	{
		skf = sscanf(rdbuff+1, "%d", &send_to_pid);
		if(skf == 1)
		{
			gammapb_irq_enabled = 1; 
		}
		else /* sscanf failed or did something else unwanted */
		{
			send_to_pid = 0; 
			gammapb_irq_enabled = 0; 
		}
	}
	else
	{
		printk(KERN_INFO "%s(%d): Written unknown command %s\n", THISFILE, __LINE__, rdbuff); 
	}
	return(count); 
}



